#ifndef _MY_UART_
#define _MY_UART_
#include "pinmacro.h"

//   UART (    ATmega8 / 48 / 88 / ...)
#ifndef UART_RX	
	#define UART_RX	D,0
#endif
#ifndef UART_TX
	#define UART_TX	D,1
#endif
// ,    (  2^5 = 32 )
#ifndef UART_BUF_PWR
  #define UART_BUF_PWR (5)
#endif

#ifdef UART_INTERRUPTS
  #include <avr/interrupt.h>
#endif
//    
#ifndef UBRRH
#define UBRRH UBRR0H
#endif
#ifndef UBRRL
#define UBRRL UBRR0L
#endif
#ifndef UCSRA
#define UCSRA UCSR0A
#endif
#ifndef UCSRB
#define UCSRB UCSR0B
#endif
#ifndef UCSRC
#define UCSRC UCSR0C
#endif
#ifndef U2X
#define U2X U2X0
#endif
#ifndef RXEN
#define RXEN RXEN0
#endif
#ifndef TXEN
#define TXEN TXEN0
#endif
#ifndef UCSZ1
#define UCSZ1 UCSZ01
#endif
#ifndef UCSZ0
#define UCSZ0 UCSZ00
#endif
#ifndef RXC
#define RXC RXC0
#endif
#ifndef TXC
#define TXC TXC0
#endif
#ifndef UDR
#define UDR UDR0
#endif
#ifndef UDRE
#define UDRE UDRE0
#endif

#define UART_SIZE (1<<UART_BUF_PWR)
#define UART_MASK (UART_SIZE - 1)
//  
typedef struct{
  unsigned char st,en;
  unsigned char arr[UART_SIZE];
}uart_buffer;

uart_buffer uart_rx;
uart_buffer uart_tx;

#define CALC_UBRR(baud) ((F_CPU/4/(baud)-1)/2)

/*   UART ()
 *  :
 *   speed -  ()
 */
#define uart_speed(speed) do{UBRRH = (CALC_UBRR(speed)>>8)&0x0F; UBRRL = CALC_UBRR(speed)&0xFF;}while(0)

/*  UART      
 *      UART_INTERRUPTS
 */
void uart_init(){
  DDR_1(UART_TX);
  DDR_0(UART_RX);
  PORT_0(UART_TX);
  PORT_0(UART_RX);
  UCSRA=(1<<U2X);
#ifdef UART_INTERRUPTS
  UCSRB=(1<<RXEN | 1<<TXEN | 1<<RXCIE);
#else
  UCSRB=(1<<RXEN | 1<<TXEN);
#endif
#ifdef URSEL
  UCSRC=(1<<URSEL | 1<<UCSZ1 | 1<<UCSZ0);
#else
  UCSRC=(1<<UCSZ1 | 1<<UCSZ0);
#endif
  uart_rx.st=0; uart_rx.en=0; uart_tx.st=0; uart_tx.en=0;
}

/*     
 */
inline unsigned char uart_rx_size(){
  return ((uart_rx.st - uart_rx.en) & UART_MASK);
}

/*       
 */
inline unsigned char uart_tx_size(){
  return ((uart_tx.st - uart_tx.en) & UART_MASK);
}

/* (  !)  1    
 */
char uart_rx_read(){
  char res;
  if(uart_rx_size()==0)return 0;
  res = uart_rx.arr[uart_rx.st++];
  uart_rx.st &= UART_MASK;
  return res;
}

/* (  !)  1    
 */
char uart_tx_read(){
  char res;
  if(uart_tx_size()==0)return 0;
  res = uart_tx.arr[uart_tx.st++];
  uart_tx.st &= UART_MASK;
  return res;
}

/* (  !)  1    
 */
void uart_rx_write(char data){
  if(uart_rx_size()!=1){
    uart_rx.arr[uart_rx.en++] = data;
    uart_rx.en &= UART_MASK;
  }
}

/* (  !)  1    
 */
void uart_tx_write(char data){
  if(uart_tx_size()!=1){
    uart_tx.arr[uart_tx.en++] = data;
    uart_tx.en &= UART_MASK;
  }
}

/*     ()
 */
#define uart_is_empty()	(uart_rx_size()==0)

/*     UART
 */
#define uart_read()	uart_rx_read()

/*     UART    
 */
char uart_scan(){
  return uart_rx.arr[uart_rx.st];
}

/*    UART
 */
#ifdef UART_INTERRUPTS
void uart_write(char data){
  asm("cli");
  if((uart_tx_size()==0) && (!(UCSRB & (1<<UDRIE)))){
    UDR=data;
    UCSRB |= (1<<UDRIE);
  }else uart_tx_write(data);
  asm("sei");
}
#else
inline void uart_write(char data){
  uart_tx_write(data);
}
#endif

#ifdef UART_INTERRUPTS
//-----------------------------------------------------------------
ISR(USART_RXC_vect){
  char temp=UDR;
  uart_rx_write(temp);
}
ISR(USART_UDRE_vect){
  if(uart_tx_size()!=0)UDR=uart_tx_read();else UCSRB&=~(1<<UDRIE);
}
#else
//--- ----------------------------------------------------------
/*          
 */
inline void uart_update(){
  if( UCSRA & (1<<RXC) ){
    uart_rx_write(UDR);
  }else if( UCSRA & (1<<UDRE) ){
    if(uart_tx_size()!=0)UDR = uart_tx_read();
  }
}
#endif

#endif
